import unittest
from unittest._log import _LoggingWatcher
import sys
import time

import logging

# pylint: disable=no-name-in-module
from log_indented import logged, log_info, LoggedBlock  # type: ignore

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


def f_level_3() -> None:
    log_info("level 3: enter")
    f_level_4_with()
    log_info("level 3: exit")


def f_level_4_with() -> None:
    with LoggedBlock("with logged block level 4", logger):
        log_info("inside logged block level 4")
        logger.info("just a regular log")


@logged(logger)
def count_chicken() -> int:
    return 3


@logged(logger)
def count_ducks() -> int:
    return 7


@logged(logger)
def count_birds() -> int:
    return count_chicken() + count_ducks()


@logged(logger)
def count_goats() -> int:
    return 7


@logged(logger)
def count_sheep() -> int:
    return 0


@logged(logger)
def count_barnyard_animinals() -> int:
    total_animal_count: int = count_birds() + count_goats() + count_sheep()
    log_info(f"total barnyard animals: {total_animal_count}")
    return total_animal_count


@logged(logger)
def compute_the_answer() -> int:
    for i in range(10):
        time.sleep(0.1)
        log_info(f"{i}, computing the answer")
    return 42


class TestLogIndented(unittest.TestCase):
    @logged(logger)
    def _f_level_2(self) -> None:
        log_info("level 2")
        f_level_3()
        # pylint: disable=redundant-unittest-assert
        self.assertTrue(True)

    @logged(logger)
    def _f_level_1(self) -> None:
        log_info("level 1")
        self._f_level_2()
        # pylint: disable=redundant-unittest-assert
        self.assertTrue(True)

    @logged(logger)
    def test_basic(self) -> None:
        with self.assertLogs() as captured:
            self._f_level_1()

        self._validate_captured_logs(
            expected_lines=[
                "+ TestLogIndented._f_level_1: enter",
                "  TestLogIndented._f_level_1: level 1",
                "    + TestLogIndented._f_level_2: enter",
                "      TestLogIndented._f_level_2: level 2",
                "      TestLogIndented._f_level_2: level 3: enter",
                "        + with logged block level 4: enter",
                "          with logged block level 4: inside logged block level 4",
                "just a regular log",
                "        - with logged block level 4: exit. took ",
                "      TestLogIndented._f_level_2: level 3: exit",
                "    - TestLogIndented._f_level_2: exit. took ",
                "- TestLogIndented._f_level_1: exit. took ",
            ],
            captured=captured,
        )

    @logged(logger)
    def test_count_animals(self) -> None:
        with self.assertLogs() as captured:
            animal_count: int = count_barnyard_animinals()
            self.assertEqual(animal_count, 17)
        self._validate_captured_logs(
            expected_lines=[
                "    + count_barnyard_animinals: enter",
                "        + count_birds: enter",
                "            + count_chicken: enter",
                "            - count_chicken: exit. took ",
                "            + count_ducks: enter",
                "            - count_ducks: exit. took ",
                "        - count_birds: exit. took ",
                "        + count_goats: enter",
                "        - count_goats: exit. took ",
                "        + count_sheep: enter",
                "        - count_sheep: exit. took ",
                "      count_barnyard_animinals: total barnyard animals: 17",
                "    - count_barnyard_animinals: exit. took ",
            ],
            captured=captured,
        )

    @logged(logger)
    def test_important_computation(self) -> None:
        with self.assertLogs() as captured:
            the_answer: int = compute_the_answer()
            self.assertEqual(the_answer, 42)

        self._validate_captured_logs(
            expected_lines=[
                "    + compute_the_answer: enter",
                "      compute_the_answer: 0, computing the answer",
                "      compute_the_answer: 1, computing the answer",
                "      compute_the_answer: 2, computing the answer",
                "      compute_the_answer: 3, computing the answer",
                "      compute_the_answer: 4, computing the answer",
                "      compute_the_answer: 5, computing the answer",
                "      compute_the_answer: 6, computing the answer",
                "      compute_the_answer: 7, computing the answer",
                "      compute_the_answer: 8, computing the answer",
                "      compute_the_answer: 9, computing the answer",
                "    - compute_the_answer: exit. took ",
            ],
            captured=captured,
        )

        # self._validate_captured_logs(expected_lines, captured)

    def _validate_captured_logs(self, expected_lines: list[str], captured: _LoggingWatcher) -> None:
        self.assertEqual(len(captured.records), len(expected_lines))
        for index, expected_string in enumerate(expected_lines):
            self.assertIn(expected_string, captured.records[index].getMessage())

    def setUp(self) -> None:
        self.stream_handler = logging.StreamHandler(sys.stdout)
        logger.addHandler(self.stream_handler)

    def tearDown(self) -> None:
        logger.removeHandler(self.stream_handler)



"""
Basic logging module.
"""
from typing import Any, Callable, cast, TypeVar, Literal, Optional, Tuple, Type
import time
from functools import wraps
import logging
import threading
from types import TracebackType

default_logger: logging.Logger = logging.getLogger(__name__)

PREFIX_ENTER: str = "+ "
PREFIX_EXIT: str = "- "
PREFIX_MESSAGE: str = "  "


class Indentations:
    def __init__(self) -> None:
        self.indents: dict[int, list[Tuple[logging.Logger, str]]] = {}
        self.lock: threading.Lock = threading.Lock()

    def push_logger(self, logger: logging.Logger, name: str) -> None:
        thread_id: int = threading.get_ident()
        with self.lock:
            if thread_id not in self.indents:
                self.indents[thread_id] = []
            self.indents[thread_id].append((logger, name))

    def pop_logger(self) -> None:
        thread_id: int = threading.get_ident()
        with self.lock:
            if thread_id not in self.indents:
                # This points to a likely bug in this class. Log the error, but don't disrupt business logic.
                default_logger.warning("trying to unindent on an untracked thread, %s", thread_id)
                return

            self.indents[thread_id].pop()

            # If we are no longer tracking indentations on this thread, stop tracking, to avoid unnecessarily growing the dictionary.
            if not self.indents[thread_id]:
                del self.indents[thread_id]

    def get_indent_and_logger(self) -> Tuple[int, logging.Logger, str]:
        thread_id: int = threading.get_ident()
        with self.lock:
            if thread_id not in self.indents:
                return 0, default_logger, ""
            logger_to_use, name = self.indents[thread_id][-1]
            indent: int = len(self.indents[thread_id]) - 1
            return indent, logger_to_use, name


class IndentedLog:

    indentations: Indentations = Indentations()

    @staticmethod
    def prepare_compose_log(with_prefix: bool, *args: Any) -> Tuple[str, logging.Logger]:
        indent_level, logger_to_user, name = IndentedLog.indentations.get_indent_and_logger()

        string_to_log: str = PREFIX_MESSAGE if with_prefix else ""
        string_to_log += "  " * (indent_level * 2)
        if with_prefix:
            string_to_log += name + ": "
        for argument in args:
            string_to_log += f"{argument}"

        return string_to_log, logger_to_user

    @staticmethod
    def push_logger(slogger: logging.Logger, name: str) -> None:
        IndentedLog.indentations.push_logger(slogger, name)

    @staticmethod
    def info(with_prefix: bool, *args: Any) -> None:
        log_string, logger_to_use = IndentedLog.prepare_compose_log(with_prefix, *args)
        logger_to_use.info(log_string)

    @staticmethod
    def warning(with_prefix: bool, *args: Any) -> None:
        log_string, logger_to_use = IndentedLog.prepare_compose_log(with_prefix, *args)
        logger_to_use.warning(log_string)

    @staticmethod
    def error(with_prefix: bool, *args: Any) -> None:
        log_string, logger_to_use = IndentedLog.prepare_compose_log(with_prefix, *args)
        logger_to_use.error(log_string)


def log_info(*args: Any) -> None:
    IndentedLog.info(True, *args)


def log_warn(*args: Any) -> None:
    IndentedLog.warning(True, *args)


def log_error(*args: Any) -> None:
    IndentedLog.error(True, *args)


class LoggedBlock:
    def __init__(self, name: str, flogger: logging.Logger = default_logger) -> None:
        self.start_time: float = 0
        self.name: str = name
        self.logger: logging.Logger = flogger

    def __enter__(self):  # type: ignore
        IndentedLog.indentations.push_logger(self.logger, self.name)
        IndentedLog.info(False, f"{PREFIX_ENTER}{self.name}: enter")
        # self.logger.info(self.name + ": enter")
        self.start_time = int(time.time())
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> Literal[False]:
        duration: float = self.since_start() * 1000.0
        exception_text: str = f"exception: '{exc_type}' - '{exc_value}'" if exc_value else ""
        message_str: str = f"{PREFIX_EXIT}{self.name}: exit. took {duration:,.1f} ms. {exception_text}"
        if exc_value:
            IndentedLog.info(False, message_str)
        else:
            IndentedLog.info(False, message_str)
        IndentedLog.indentations.pop_logger()
        return False

    def since_start(self) -> float:
        return time.time() - self.start_time


# pylint: disable=invalid-name
F = TypeVar("F", bound=Callable[..., Any])


def logged(flogger: logging.Logger = default_logger) -> Callable[[F], F]:
    def decorate(function: F) -> F:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """Logs the timing for a function."""
            with LoggedBlock(function.__qualname__, flogger):
                ret = function(*args, **kwargs)
            return ret

        return cast(F, wrapper)

    return decorate
